/* Copyright 2007, 2008 DJ Delorie <dj@delorie.com>
   Distributed under the terms of the GNU General Public License
   version 3, see COPYING-3.0 for details.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include "toc.h"

#define EE_PORT 1003
#define EE_HOST "alarmclock"

#define NUM_TOC_ENTRIES TOC_MAX
#define NUM_FONT_ENTRIES 15
#define MAX_ALARMS  20
#define MAX_WEBS  10
#define MAX_STRINGS  40

typedef unsigned char byte;
typedef struct {
  byte l;
  byte h;
} word;

int ee;

typedef struct {
  word addr;
  word size;
} TOCentry;

typedef struct {
  void *data;
  int size;
  int addr;
} Chunk;

typedef struct {
  byte hours_from_utc;
  byte vol_ramp;
  byte sleep_volume;
  char rtc_calib;
  char disp_24hr;
} CONFIGtype;

typedef struct {
  word mp3;
  word oled;
  word name;
} WEBdest;

typedef struct {
  int addr;
  Chunk *mp3;
  Chunk *oled;
  Chunk *name;
} LWEBdest;

typedef struct {
  byte hour;
  byte minute;
  byte weekdays;
  byte month;
  byte day;
  byte zen;
  word web;
} ALARMentry;

typedef struct {
  int hour;
  int minute;
  int weekdays;
  int month;
  int day;
  int zen;
  LWEBdest *web;
} LALARMentry;

typedef struct {
  word glyph[NUM_FONT_ENTRIES];
} FONTtype;

typedef struct {
  void *glyph;
  int size;
  int addr;
} LGLYPH;

typedef struct {
  int width;
  int height;
  byte **pix;
} PGM;

int no_download = 0;
int errors = 0;
FILE *cf;
TOCentry TOC[NUM_TOC_ENTRIES];
Chunk LTOC[NUM_TOC_ENTRIES];
CONFIGtype config;
ALARMentry alarms[MAX_ALARMS];
LALARMentry lalarms[MAX_ALARMS];
WEBdest webs[MAX_WEBS];
LWEBdest lwebs[MAX_WEBS];
Chunk strings[MAX_STRINGS];
FONTtype main_font;
LGLYPH lmain_font[NUM_FONT_ENTRIES];
FONTtype night_font;
LGLYPH lnight_font[NUM_FONT_ENTRIES];
FONTtype half_font;
LGLYPH lhalf_font[NUM_FONT_ENTRIES];
FONTtype mini_font;
LGLYPH lmini_font[NUM_FONT_ENTRIES];
Chunk console_font;
Chunk beep;
Chunk silence;
Chunk images[NUM_IMAGES];

const char *
toc_name(int i)
{
  if (i < TOC_images)
    return toc_names[i];
  i -= TOC_images;
  if (i < sizeof(image_names)/sizeof(image_names[0])
      && image_names[i])
    return image_names[i];
  return "unk";
}

int myfgetc(FILE *f)
{
  int rv = fgetc (f);
  printf("[%c]", rv);
  return rv;
}
/*#define fgetc myfgetc*/

int
read_pgm_int (FILE *f)
{
  int c, rv = 0;
  while ((c = fgetc (f)) != EOF)
    {
      if (c == '#')
	do { c = fgetc (f); } while (c != '\n');
      if (isdigit (c))
	{
	  rv = c - '0';
	  while ((c = fgetc(f)) != EOF)
	    {
	      if (!isdigit (c))
		return rv;
	      rv *= 10;
	      rv += c - '0';
	    }
	  return rv;
	}
    }
}

PGM *
load_pgm (char *filename)
{
  int w, h, d;
  int i, x, y;
  FILE *f;
  PGM *pgm;

  f = fopen(filename, "rb");
  if (!f)
    {
      perror(filename);
      exit(1);
    }
  if (fgetc (f) != 'P'
      || fgetc (f) != '5')
    {
      printf("%s not a binary PGM file\n", filename);
      exit(1);
    }
  w = read_pgm_int (f);
  h = read_pgm_int (f);
  d = read_pgm_int (f);
  pgm = (PGM *)malloc (sizeof(PGM));
  pgm->width = w;
  pgm->height = h;
  pgm->pix = (byte **)malloc(w * sizeof (byte *));
  for (x=0; x<w; x++)
    pgm->pix[x] = (byte *)malloc(h);
  for (y=0; y<h; y++)
    for (x=0; x<w; x++)
      pgm->pix[x][y] = fgetc(f);
  fclose(f);
  return pgm;
}

void
connect_to_alarm_clock ()
{
  struct sockaddr_in addr;
  struct hostent *he;
  int one = 1;

  ee = socket (PF_INET, SOCK_STREAM, 0);
  setsockopt (ee, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

  addr.sin_family = AF_INET;
  addr.sin_port = htons(EE_PORT);

  he = gethostbyname (EE_HOST);
  if (!he)
    {
      fprintf(stderr, "msg: unknown host `%s'\n", EE_HOST);
      exit(1);
    }
  memcpy (&addr.sin_addr.s_addr, he->h_addr_list[0], he->h_length);

  if (connect (ee, (const struct sockaddr *)&addr, sizeof(addr)) != 0)
    {
      perror("connect");
      exit(1);
    }
  setsockopt (ee, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
  setsockopt (ee, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));

  setbuf (stdout, 0);
}

static void
WD (unsigned char *buf, int count)
{
  int i;
  for (i=0; i<count; i++)
    printf("W: %02x\n", buf[i]);
}

static void
RD (unsigned char *buf, int count)
{
  int i;
  for (i=0; i<count; i++)
    printf("R:    %02x\n", buf[i]);
}

#define WD(a,b)
#define RD(a,b)

eeread(unsigned int addr, unsigned char *buf, int count)
{
  unsigned char abuf[4];
  while (count > 0)
    {
      int c = count;
      if (c > 255)
	c = 255;
      abuf[0] = 'R';
      abuf[1] = addr & 0xff;
      abuf[2] = addr >> 8;
      abuf[3] = c;
      WD (abuf, 4);
      write (ee, abuf, 4);
      read (ee, buf, c);
      RD (buf, c);
      buf += c;
      count -= c;
      addr += c;
    }
}

eewrite(unsigned int addr, unsigned char *buf, int count)
{
  int i;
  unsigned char abuf[4];
  struct iovec io[2];

  io[0].iov_base = abuf;
  io[0].iov_len = 4;

  while (count > 0)
    {
      int c = count;
      if (c > 255)
	c = 255;
      abuf[0] = 'W';
      abuf[1] = addr & 0xff;
      abuf[2] = addr >> 8;
      abuf[3] = c;
      WD(abuf, 4);
      WD(buf, c);
      io[1].iov_base = buf;
      io[1].iov_len = c;
      writev (ee, io, 2);
      read (ee, abuf, 1);
      RD (abuf, 1);
      buf += c;
      count -= c;
      addr += c;
      putchar('.');
    }
}

unsigned char buf[65536];
char *line[30];

usage ()
{
  fprintf(stderr, "usage: eetool -r <addr> <count>\n");
  fprintf(stderr, "       eetool -w <addr> [bytes . . .]\n");
  fprintf(stderr, "       eetool\n");
  exit(1);
}

int
getline ()
{
  int i;
  char *lp;

  if (fgets (buf, sizeof(buf), cf) == NULL)
    return 0;
  if (strlen(buf) > 0 && buf[strlen(buf)-1] == '\n')
    buf[strlen(buf)-1] == '\0';
  if (strlen(buf) > 0 && buf[strlen(buf)-1] == '\r')
    buf[strlen(buf)-1] == '\0';

  i = 0;
  lp = (char *)buf;
  memset(line, 0, sizeof(line));

  while (1)
    {
      while (*lp && isspace (*lp))
	lp++;
      if (*lp == 0)
	return 1;
      if (*lp == '"')
	{
	  line[i++] = ++lp;
	  while (*lp && *lp != '"')
	    lp++;
	}
      else
	{
	  line[i++] = lp;
	  while (*lp && !isspace(*lp))
	    lp++;
	}
      if (*lp)
	*lp++ = 0;
      else
	return 1;
    }
}

/*=============================================================================*/

static void
load_config ()
{
  config.hours_from_utc = line[1] ? atoi (line[1]) : 0;
  config.vol_ramp = line[2] ? atoi (line[2]) : 0;
  config.sleep_volume = line[3] ? atoi (line[3]) : 128;
  config.rtc_calib = line[4] ? atoi (line[4]) : 0;
  config.disp_24hr = line[5] ? atoi (line[5]) : 0;
}

static void
load_alarm ()
{
  int active;
  int which;
  LALARMentry *la;

  if (!line[6])
    {
      fprintf (stderr, "usage: alarm <which> <hh:mm> <SuMoTuWeThFrSa|*> <month> <day> <zen> <web>\n");
      errors ++;
      return;
    }
  which = atoi(line[1]);
  la = lalarms + which;

  sscanf (line[2], "%d:%d", &la->hour, &la->minute);
  la->weekdays = 0;
  active = 1;
  while (line[3][0])
    {
      if (strncmp (line[3], "No", 2) == 0) active = 0;
      if (strncmp (line[3], "Su", 2) == 0) la->weekdays |= 0x40;
      if (strncmp (line[3], "Mo", 2) == 0) la->weekdays |= 0x20;
      if (strncmp (line[3], "Tu", 2) == 0) la->weekdays |= 0x10;
      if (strncmp (line[3], "We", 2) == 0) la->weekdays |= 0x08;
      if (strncmp (line[3], "Th", 2) == 0) la->weekdays |= 0x04;
      if (strncmp (line[3], "Fr", 2) == 0) la->weekdays |= 0x02;
      if (strncmp (line[3], "Sa", 2) == 0) la->weekdays |= 0x01;
      if (strcmp (line[3], "*") == 0) { la->weekdays |= 0x7f; break; }
      line[3] += 2;
    }
  if (la->weekdays && active)
    la->weekdays |= 0x80;

  la->month = atoi(line[4]);
  la->day = atoi(line[5]);
  la->zen = atoi(line[6]);
  if (strcmp (line[7], "-") == 0)
    la->web = NULL;
  else
    la->web = lwebs + atoi (line[7]);
}

static Chunk *
store_string (char *str)
{
  int i;
  for (i=0; i<MAX_STRINGS; i++)
    if (strings[i].data
	&& strcmp (strings[i].data, str) == 0)
      return strings + i;
  for (i=0; i<MAX_STRINGS; i++)
    if (!strings[i].data)
      {
	strings[i].data = strdup (str);
	strings[i].size = strlen (str);
	return strings + i;
      }
}

static void
load_web ()
{
  int which;
  LWEBdest *web;

  which = atoi (line[1]);
  web = lwebs + which;

  if (strcmp (line[2], "mp3") == 0)
    web->mp3 = store_string(line[3]);
  if (strcmp (line[2], "oled") == 0)
    web->oled = store_string(line[3]);
  if (strcmp (line[2], "name") == 0)
    web->name = store_string(line[3]);
}

color(int s)
{
  static int i = 0;
  i = 2-i;
  printf("\033[%dm", i + (s ? 31 : 34));
}

static byte *oraw;

static int
type_A_byte (byte *raw)
{
  int rv=0, i;
  for (i=0; i<8; i++)
    if (raw[i] > 0x80)
      rv |= (1<<i);
  return rv;
}

static int
type_A_seq (byte *raw, int rc)
{
  int seq = 1, b0;
  printf("type_A_seq %d %d... ", raw-oraw, rc);
  b0 = type_A_byte (raw);
  while (seq < rc/8
	 && type_A_byte (raw+seq*8) == b0)
    seq ++;
  if (seq > 127)
    seq = 127;
  printf("seq %d\n", seq);
  return seq;
}

static int
type_A_nonseq (byte *raw, int rc, int seq)
{
  int rv = 1;
  printf("type_A_nonseq %d %d, %d... ", raw-oraw, rc, seq);
  while (rc > 8)
    {
      if (type_A_seq (raw+8, rc) >= seq)
	break;
      raw += 8;
      rc -= 8;
      rv ++;
    }
  if (rv > 127)
    rv = 127;
  printf("rv %d\n", rv);
  return rv;
}

/* RAW is a stream of RC bytes, each of which is 0..0xff.  RLL is filled
   with the compressed stream, and its length is returned.  */
static int
type_A_compress (byte *raw, int rc, byte *rll)
{
  byte *rll0 = rll;
  printf("\ncompress %d\n", rc);
  oraw = raw;
  while (rc > 0)
    {
      int seq = type_A_seq (raw, rc);
      if (seq > 1)
	{
	  *rll++ = 0x80 | seq;
	  *rll++ = type_A_byte (raw);
	  raw += seq*8;
	  rc -= seq*8;
	}
      else
	{
	  int b;
	  seq = type_A_nonseq (raw, rc, 3);
	  *rll++ = seq;
	  for (b=0; b<seq; b++)
	    {
	      *rll++ = type_A_byte (raw);
	      raw += 8;
	      rc -= 8;
	    }
	}
    }
  *rll++ = 0;
  return rll - rll0;
}

static int
type_B_seq (byte *raw, int rc)
{
  int seq = 2;
  while (seq < rc
	 && raw[0] == raw[seq]
	 && raw[1] == raw[seq+1])
    seq += 2;
  seq /= 2;
  if (seq > 15)
    seq = 15;
  return seq;
}

static int
type_B_nonseq (byte *raw, int rc, int seq)
{
  int rv = 1;
  while (rc > 3)
    {
      if (type_B_seq (raw+rv*4, rc) >= seq)
	break;
      raw += 4;
      rc -= 4;
      rv ++;
    }
  if (rv > 15)
    rv = 15;
  return rv;
}

/* RAW is a stream of RC bytes, each of which is 0..3.  RLL is filled
   with the compressed stream, and its length is returned.  */
static int
type_B_compress (byte *raw, int rc, byte *rll)
{
  byte *rll0 = rll;
  while (rc > 3)
    {
      int seq = type_B_seq (raw, rc);
      if (seq > 1)
	{
	  *rll++ = seq << 4 | raw[0]*4 | raw[1];
	  raw += seq*2;
	  rc -= seq*2;
	}
      else
	{
	  int b;
	  seq = type_B_nonseq (raw, rc, 2);
	  *rll++ = seq;
	  for (b=0; b<seq; b++)
	    {
	      *rll++ = (raw[0] << 6
			| raw[1] << 4
			| raw[2] << 2
			| raw[3]);
	      raw += 4;
	      rc -= 4;
	    }
	}
    }

  /* stragglers */
  while (rc)
    {
      *rll++ = 1 << 4 | raw[0]*4 | raw[1];
      raw += 2;
      rc -= 2;
    }

  *rll++ = 0;
  return rll - rll0;
}

static int
type_C_seq (byte *raw, int rc)
{
  int seq = 1;
  while (seq < rc
	 && raw[0] == raw[seq])
    seq ++;
  seq >>= 1;
  if (seq > 15)
    seq = 15;
  return seq;
}

static int
type_C_nonseq (byte *raw, int rc, int seq)
{
  int rv = 1;
  while (rc>2)
    {
      if (rc > 3)
	if (type_C_seq (raw+2, rc) >= seq)
	  break;
      raw += 2;
      rc -= 2;
      rv ++;
    }
  if (rv > 15)
    rv = 15;
  return rv;
}

static int djwidth, col;
static byte *rllp;
static void
djput (int c, byte *rll)
{
  static char map[] = ".123456789abcdef";
  putchar(map[c]);
  if (++col >= djwidth)
    {
      putchar('\t');
      while (rllp < rll)
	printf(" %02x", *rllp++);
      putchar('\n');
      col = 0;
    }
}

static int
type_C_compress (byte *raw, int rc, byte *rll)
{
  int i;
  byte *rll0 = rll;
  rllp = rll;
  while (rc > 0)
    {
      int seq = type_C_seq (raw, rc);
      if (seq > 0)
	{
	  *rll++ = (seq << 4) | raw[0];
#if 0
	  color(1);
	  for (i=0; i<seq*2; i++)
	    djput(raw[i], rll);
#endif
	  raw += seq*2;
	  rc -= seq*2;
	}
      else
	{
	  int b;
	  seq = type_C_nonseq (raw, rc, 2);
#if 0
	  color(0);
#endif
	  *rll++ = seq;
	  for (b=0; b<seq; b++)
	    {
	      *rll++ = (raw[0] << 4
			| raw[1]);
#if 0
	      djput(raw[0], rll);
	      djput(raw[1], rll);
#endif
	      raw += 2;
	      rc -= 2;
	    }
	}
    }
  *rll++ = 0;
  return rll - rll0;
}

int oldshadows[3][3] = {
  { 10,  10,   1 },
  { 10,   0, -10 },
  { -1, -10, -10 },
};

int shadows[3][3] = {
  { 0,  10,   0 },
  { 10,   0, -11 },
  { 0, -11, 0 },
};

static void
load_font (int t, FONTtype *f, LGLYPH *lg, char compression_type, int time_width)
{
  int x, y, g;
  int wi = 0;
  PGM *pgm;
  int x0[NUM_FONT_ENTRIES];
  int w[NUM_FONT_ENTRIES];
  byte *rawpix;
  byte *rllpix;
  int li = 1;

  if (compression_type == 0)
    {
      compression_type = toupper (line[1][0]);
      li ++;
    }

  printf("\nFont: %s\n", line[li]);
  pgm = load_pgm (line[li]);
  g = wi = 0;
  for (x=0; x<pgm->width; x++)
    {
      int s = 0;
      for (y=0; y<pgm->height; y++)
	s += pgm->pix[x][y] > 10 ? 1 : 0;
      if (s)
	{
	  if (wi == 0)
	    {
	      x0[g] = x-1;
	      wi ++;
	    }
	  wi++;
	}
      else if (wi)
	{
	  w[g] = wi + 1;
	  if (w[g] & 1)
	    w[g]++;
	  wi = 0;
	  g++;
	  if (g == NUM_FONT_ENTRIES)
	    break;
	}
    }
#if 1
  for (x=0; x<g; x++)
    printf("%2d: %3d - %3d\n", x, x0[x], w[x]);
#endif
  if (g < NUM_FONT_ENTRIES)
    {
      fprintf(stderr, "%s doesn't have enough glyphs\n", line[li]);
      errors ++;
      return;
    }

  LTOC[t].size = sizeof(FONTtype);

  for (g=0; g<NUM_FONT_ENTRIES; g++)
    {
      byte *rp;
      int rllen, rlp;

      rawpix = (byte *) malloc (w[g] * pgm->height);
      rllpix = (byte *) malloc (w[g] * pgm->height);
      rp = rawpix;
      rlp = w[g] * pgm->height / 2;

      if (compression_type == 'B')
	{
	  for (y=0; y<pgm->height; y++)
	    for (x = x0[g]; x < x0[g]+w[g]; x++)
	      {
		if (pgm->pix[x][y] == 0)
		  {
		    *rp++ = 0;
		  }
		else
		  {
		    int sx, sy;
		    int s = 0;
		    for (sx=0; sx<3; sx++)
		      for (sy=0; sy<3; sy++)
			if (pgm->pix[x+sx-1][y+sy-1] == 0)
			  s += shadows[sx][sy];
		    if (s > 0)
		      *rp++ = 2;
		    else if (s < 0)
		      *rp++ = 1;
		    else
		      *rp++ = 3;
		  }
	      }
	}
      if (compression_type == 'C')
	{
	  for (y=0; y<pgm->height; y++)
	    for (x = x0[g]; x < x0[g]+w[g]; x++)
	      *rp++ = (pgm->pix[x][y] >> 4) & 0x0f;
	}

      rllpix[0] = compression_type;
      rllpix[1] = w[g];
      djwidth = w[g];
      col = 0;
      rllpix[2] = pgm->height;
      switch (compression_type)
	{
	case 'A':
	  break;
	case 'B':
	  printf("\ncompress glyph %d\n", g);
	  rllen = type_B_compress (rawpix, rp-rawpix, rllpix+3);
	  //printf("B: %d -> %d bytes (%d)\n", rp-rawpix, rllen, rlp);
	  break;
	case 'C':
	  rllen = type_C_compress (rawpix, rp-rawpix, rllpix+3);
	  //printf("C: %d -> %d bytes (%d)\n", rp-rawpix, rllen, rlp);
	  break;
	}
      rllen += 3;

      lg[g].glyph = rllpix;
      lg[g].size = rllen;
      lg[g].addr = 0;

      LTOC[t].size += rllen;
    }
  {
    int fw=0, i, c;
    static char mins[] = "10:00";
    static char maxs[] = "12;59";
    for (i=0; i<5; i++)
      {
	int wm=0;
	for (c=mins[i]-'0'; c<=maxs[i]-'0'; c++)
	  if (wm < w[c])
	    wm = w[c];
	fw += wm;
      }
    printf("font width: %d\n", fw);
    if (fw > time_width)
      {
	printf("... this exceeds the limit of %d\n", time_width);
	errors ++;
      }
  }
}

static void
load_console_font ()
{
  PGM *pgm;
  byte *d;
  int c, cx, cy, x, y, b, bc;

  pgm = load_pgm (line[1]);

  d = (byte *) malloc (128 * 6);
  console_font.data = d;
  console_font.size = 128*6;

  for (c=0; c<128; c++)
    {
      cx = (c % 16) * 6;
      cy = (c / 16) * 8;
      bc = 0;
      b = 0;
      for (y=0; y<8; y++)
	for (x=0; x<6; x++)
	  {
	    bc++;
	    b >>= 1;
	    b |= pgm->pix[cx+x][cy+y] < 127 ? 0x80 : 0;
	    if (bc == 8)
	      {
		*d++ = b;
		b = 0;
		bc = 0;
	      }
	  }
    }
}

static void
load_chunk (Chunk *c)
{
  struct stat s;
  FILE *f;

  if (stat(line[1], &s) < 0)
    {
      perror(line[1]);
      exit(1);
    }
  c->data = (byte *)malloc(s.st_size);
  f = fopen(line[1], "rb");
  fread (c->data, 1, s.st_size, f);
  fclose(f);
  c->size = s.st_size;
}

static void
dump_A (byte *r, int w, int h)
{
  int x = 0, i, b, bit, cc=0;
  byte *or = r;

  printf("%4d: ", 0);
  while (1)
    {
      b = *r++;
      color(0);
      if (b & 0x80)
	{
	  color(0);
	  i = *r++;
	  b &= 0x7f;
	  while (b--)
	    {
	      for (bit=0; bit<8; bit++)
		{
		  if (x == w)
		    {
		      x = 0;
		      putchar('\n');
		      printf("%4d: ", cc);
		    }
		  putchar((i & (1<<bit)) ? '#' : '.');
		  cc++;
		  x++;
		}
	    }
	}
      else if (b)
	{
	  color(1);
	  while (b--)
	    {
	      i = *r++;
	      for (bit=0; bit<8; bit++)
		{
		  if (x == w)
		    {
		      x = 0;
		      putchar('\n');
		      printf("%4d: ", cc);
		    }
		  putchar((i & (1<<bit)) ? '#' : '.');
		  cc++;
		  x++;
		}
	    }
	}
      else
	break;
    }
  printf("\033[0m\n");
  for (; or<r;)
    {
      printf(" %02x", *or++);
    }
  printf("\n");
}

static void
load_image ()
{
  struct stat s;
  FILE *f;
  int idx, x, y, sz;
  PGM *pgm;
  byte *raw, *rll, *rp;
  char image_type;

  image_type = toupper (line[1][0]);

  if (isdigit (line[2][0]))
    {
      idx = atoi (line[2]);
      if (idx < 0 || idx >= NUM_IMAGES)
	{
	  fprintf(stderr, "image number %s out of range\n", line[1]);
	  errors ++;
	  return;
	}
    }
  else
    {
      for (idx=0; image_names[idx]; idx++)
	if (strcmp (line[2], image_names[idx]) == 0)
	  break;
      if (!image_names[idx])
	{
	  fprintf(stderr, "image number %s out of range\n", line[2]);
	  errors ++;
	  return;
	}
    }

  pgm = load_pgm (line[3]);
  raw = (byte *) malloc (pgm->width * pgm->height);
  rll = (byte *) malloc (pgm->width * pgm->height);

  rll[0] = image_type;
  rll[1] = pgm->width;
  rll[2] = pgm->height;

  rp = raw;
  printf("image %d %s", idx, line[3]);
  if (image_type == 'A')
    {
      for (y=0; y<pgm->height; y++)
	for (x=0; x<pgm->width; x++)
	  *rp++ = pgm->pix[x][y];
      sz = type_A_compress (raw, rp-raw, rll+3) + 3;
      //dump_A (rll+3, pgm->width, pgm->height);
    }
  else if (image_type == 'C')
    {
      for (y=0; y<pgm->height; y++)
	for (x=0; x<pgm->width; x++)
	  *rp++ = pgm->pix[x][y] >> 4;
      sz = type_C_compress (raw, rp-raw, rll+3) + 3;
    }
  printf(" - %d -> %d\n", rp-raw, sz);
  LTOC[TOC_images+idx].data = rll;
  LTOC[TOC_images+idx].size = sz;

}

/*========================================================================================*/

#define W(w,x) (w).l = (x)&0xff, (w).h = (x)>>8
#define CW(w,x) (w).l = (x?x->addr:0)&0xff, (w).h = (x?x->addr:0)>>8
#define BCD(x) (((x)/10)*16+((x)%10))
#define PUT(t, v) memcpy (buf+LTOC[t].addr, v, LTOC[t].size)

static void
fix_font (int t, FONTtype *f, LGLYPH *lg)
{
  int i, sz;

  sz = LTOC[t].addr + sizeof(FONTtype);
  for (i=0; i<NUM_FONT_ENTRIES; i++)
    {
      lg[i].addr = sz;
      sz += lg[i].size;
    }
}

static void
calculate_addresses ()
{
  int i, sz=0;
  int oaddr = 65536;

  LTOC[TOC_TOC].size = sizeof(TOC);
  LTOC[TOC_config].size = sizeof(config);
  LTOC[TOC_alarm].size = sizeof(alarms);
  LTOC[TOC_web].size = sizeof(webs);

  sz = 0;
  for (i=0; i<MAX_STRINGS; i++)
    sz += strings[i].size + 2;
  LTOC[TOC_string].size = sz;

  LTOC[TOC_console_font].size = console_font.size;
  LTOC[TOC_default_beep].size = beep.size;
  LTOC[TOC_mp3_silence].size = silence.size;

  sz = 0;
  for (i=0; i<TOC_MAX; i++)
    sz += LTOC[i].size;
  printf("%d bytes used, %d%%\n", sz, sz*100/65536);

#define AFTER(x) LTOC[x].addr + LTOC[x].size
  LTOC[TOC_TOC].addr = 0;
  for (i=TOC_config; i<=TOC_string; i++)
    LTOC[i].addr = LTOC[i-1].addr + LTOC[i-1].size;

  oaddr = 65535;
  for (i=TOC_images+NUM_IMAGES-1; i>=TOC_console_font; i--)
    {
      oaddr -= LTOC[i].size;
      LTOC[i].addr = oaddr;
    }

  sz = LTOC[TOC_web].addr;
  for (i=0; i<MAX_WEBS; i++)
    {
      lwebs[i].addr = sz;
      sz += sizeof(WEBdest);
    }

  sz = LTOC[TOC_string].addr;
  for (i=0; i<MAX_STRINGS; i++)
    {
      strings[i].addr = sz;
      sz += strings[i].size + 2;
    }

  fix_font (TOC_main_font, &main_font, lmain_font);
  fix_font (TOC_night_font, &night_font, lnight_font);
  fix_font (TOC_half_font, &half_font, lhalf_font);
  fix_font (TOC_mini_font, &mini_font, lmini_font);
}

static void
build_eeprom ()
{
  int i;

  memset (buf, 0, 65536);

  for (i=0; i<TOC_MAX; i++)
    {
      int b, e, j;
      char bar[34];
      memset (bar, '-', 32);
      bar[32] = 0;
      bar[33] = 0;
      b = LTOC[i].addr;
      e = LTOC[i].addr+LTOC[i].size-1;
      if (LTOC[i].size)
	for (j=b/2048; j<=e/2048; j++)
	  bar[j] = '#';
      printf("%02d %4x-%4x %5d %s %s\n", i, b, e, e-b+1, bar, toc_name(i));
      W(TOC[i].addr, LTOC[i].addr);
      W(TOC[i].size, LTOC[i].size);
    }
  memcpy (buf+0, TOC, LTOC[TOC_TOC].size);

  PUT(TOC_config, &config);

  for (i=0; i<MAX_ALARMS; i++)
    {
      alarms[i].hour = BCD(lalarms[i].hour);
      alarms[i].minute = BCD(lalarms[i].minute);
      alarms[i].weekdays = lalarms[i].weekdays;
      alarms[i].month = BCD(lalarms[i].month);
      alarms[i].day = BCD(lalarms[i].day);
      alarms[i].zen = lalarms[i].zen;
      if (lalarms[i].web)
	W(alarms[i].web, lalarms[i].web->addr);
      else
	W(alarms[i].web, 0);
    }
  PUT(TOC_alarm, alarms);

  for (i=0; i<MAX_WEBS; i++)
    {
      CW(webs[i].mp3, lwebs[i].mp3);
      CW(webs[i].oled, lwebs[i].oled);
      CW(webs[i].name, lwebs[i].name);
    }
  PUT(TOC_web, webs);

  for (i=0; i<MAX_STRINGS; i++)
    {
      buf[strings[i].addr] = strings[i].size & 0xff;
      buf[strings[i].addr+1] = (strings[i].size>>8) & 0xff;
      memcpy (buf+strings[i].addr+2, strings[i].data, strings[i].size);
    }

  for (i=0; i<NUM_FONT_ENTRIES; i++)
    {
      memcpy (buf+lmain_font[i].addr, lmain_font[i].glyph, lmain_font[i].size);
      W(main_font.glyph[i], lmain_font[i].addr);
      memcpy (buf+lnight_font[i].addr, lnight_font[i].glyph, lnight_font[i].size);
      W(night_font.glyph[i], lnight_font[i].addr);
      memcpy (buf+lhalf_font[i].addr, lhalf_font[i].glyph, lhalf_font[i].size);
      W(half_font.glyph[i], lhalf_font[i].addr);
      memcpy (buf+lmini_font[i].addr, lmini_font[i].glyph, lmini_font[i].size);
      W(mini_font.glyph[i], lmini_font[i].addr);
    }
  memcpy (buf+LTOC[TOC_main_font].addr, &main_font, sizeof(main_font));
  memcpy (buf+LTOC[TOC_night_font].addr, &night_font, sizeof(night_font));
  memcpy (buf+LTOC[TOC_half_font].addr, &half_font, sizeof(half_font));
  memcpy (buf+LTOC[TOC_mini_font].addr, &mini_font, sizeof(mini_font));

  memcpy (buf+LTOC[TOC_console_font].addr, console_font.data, console_font.size);

  memcpy (buf+LTOC[TOC_default_beep].addr, beep.data, beep.size);
  memcpy (buf+LTOC[TOC_mp3_silence].addr, silence.data, silence.size);

  for (i=0; i<NUM_IMAGES; i++)
    memcpy (buf+LTOC[TOC_images+i].addr, LTOC[TOC_images+i].data, LTOC[TOC_images+i].size);

  {
    FILE *EE;
    EE = fopen("eeprom.bin", "wb");
    fwrite(buf, 1, 65536, EE);
    fclose(EE);
  }

}

static void
write_eeprom ()
{
  int i;
  printf("writing eeprom...\n");
  for (i=0; i<NUM_TOC_ENTRIES; i++)
    {
      printf("%s: ", toc_name(i));
      eewrite (LTOC[i].addr, buf+LTOC[i].addr, LTOC[i].size);
      printf("\n");
    }
}

main(int argc, char **argv)
{
  int addr, count, i;

  if (argc > 1 && strcmp (argv[1], "-n") == 0)
    no_download = 1;

  setbuf(stdout, 0);
  if (!no_download)
    connect_to_alarm_clock ();

  if (argc > 1 && strcmp (argv[1], "-r") == 0)
    {
      if (argc < 4)
	usage();
      addr = strtoul (argv[2], 0, 0);
      count = strtoul (argv[3], 0, 0);
      eeread (addr, buf, count);
      for (i=0; i<count; i++)
	printf(" %02x", buf[i]);
      printf("\n");
      exit(0);
    }

  if (argc > 1 && strcmp (argv[1], "-w") == 0)
    {
      if (argc < 4)
	usage();
      addr = strtoul (argv[2], 0, 0);
      count = argc - 3;
      for (i=0; i<count; i++)
	buf[i] = strtoul (argv[i+3], 0, 0);
      eewrite (addr, buf, count);
      exit(0);
    }

  if (argc > 1 && strcmp (argv[1], "-tz") == 0)
    {
      struct tm ltm, gtm;
      time_t now;
      int offset, dof, ld, gd;
      unsigned char toc[4];
      char tz[1];
      unsigned int addr;

      time(&now);
      ltm = *localtime(&now);
      gtm = *gmtime(&now);
      ld = ltm.tm_year * 512 + ltm.tm_mon * 32 + ltm.tm_mday;
      gd = gtm.tm_year * 512 + gtm.tm_mon * 32 + gtm.tm_mday;
      printf("local: %2d:%02d %d\n", ltm.tm_hour, ltm.tm_min, ld);
      printf("gmt:   %2d:%02d %d\n", gtm.tm_hour, gtm.tm_min, gd);

      dof = ld - gd;
      offset = ltm.tm_hour - gtm.tm_hour;
      printf("ofs: %d dof:%d\n", offset, dof);
      if (dof < 0)
	offset -= 24;
      if (dof > 0)
	offset += 24;
      printf("offset: %d\n", offset);

      /* Read config address from TOC */
      eeread (4, buf, 4);
      addr = buf[0] + buf[1] * 256;

      tz[0] = offset;
      eewrite (addr, tz, 1);
      printf("new tz: %d\n", tz[0]);

      exit(1);
    }

  cf = fopen("eeprom.cfg", "r");
  if (!cf)
    {
      fprintf(stderr, "unable to open eeprom.cfg for reading");
      perror("the error was");
      exit(1);
    }

  while (getline())
    {
      if (!line[0])
	continue;

      if (strcmp (line[0], "config") == 0)
	load_config ();

      if (strcmp (line[0], "alarm") == 0)
	load_alarm ();

      if (strcmp (line[0], "web") == 0)
	load_web ();

      if (strcmp (line[0], "main-font") == 0)
	load_font(TOC_main_font, &main_font, lmain_font, 'B', 128);
      if (strcmp (line[0], "night-font") == 0)
	load_font(TOC_night_font, &night_font, lnight_font, 0, 128);
      if (strcmp (line[0], "half-font") == 0)
	load_font(TOC_half_font, &half_font, lhalf_font, 'C', 70);
      if (strcmp (line[0], "mini-font") == 0)
	load_font(TOC_mini_font, &mini_font, lmini_font, 'C', 42);

      if (strcmp (line[0], "console-font") == 0)
	load_console_font();

      if (strcmp (line[0], "beep") == 0)
	load_chunk(&beep);
      if (strcmp (line[0], "silence") == 0)
	load_chunk(&silence);

      if (strcmp (line[0], "image") == 0)
	load_image();
    }

  if (errors)
    exit(1);
  calculate_addresses ();
  build_eeprom ();

  if (!no_download)
    {
      write_eeprom ();
      close (ee);
    }

  return 0;
}
