/* Streamconv:
 * A simple program to strip the metadata out of a Shoutcast stream
 * and pass through the MP3 datastream.
 *
 * Inputs:       HTTP URL for the .pls entry for the Shoutcast stream,
 *               on the command line.  Streamconv will extract the 
 *               port for the Shoutcast stream from the .pls file.
 * Outputs:      MP3 stream on STDOUT
 *               MP3 metadata on STDERR
 * Options:      None required nor accepted
 * Limitations:  Handles only HTTP over IPV4, not HTTPS or IPV6
 *               Currently selects only the first offered File entry
 *               in the playlist
 * Extensions:   May add the ability later to accept the Shoutcast
 *               stream on STDIN
 *
 * The Shoutcast protocol metadata format is documented at
 * http://www.smackfu.com/stuff/programming/shoutcast.html
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef FNDELAY
#define FNDELAY 0
#endif
#ifndef O_NDELAY
#define O_NDELAY 0
#endif
#ifndef O_NONBLOCK
#define O_NONBLOCK 0
#endif

#define BUFSIZE 4096
#define min(x,y) (x<y?x:y)

typedef struct hostdata {
    char          *hostname;
    char          *path;
    int            port;
    char          *stream_host;
    int            stream_port;
    int            metaint;
    int            sock;
} HOSTDATA;


int main (int argc, char **argv);
int parse_url (char *url, struct hostdata *streamhost);
int connect_host (struct hostdata *streamhost);
int get_stream_port(struct hostdata *streamhost);
void open_stream(struct hostdata *streamhost);
void usage (void);


int main(int argc, char **argv)
{
    int ret = -1;
    char *url;
    struct hostdata streamhost;
    
    switch (argc)
    {
        case 1:
        {
	    fprintf (stderr, "Usage error:  No stream URL given\n\n");
	    usage();
	    break;
	}
	
        case 2:
	{
	    bzero(&streamhost, sizeof(struct hostdata));
	    url = strdup(argv[1]); 
	    if (parse_url(url, &streamhost))
	    {
		if (connect_host(&streamhost))
		{
		    if (get_stream_port(&streamhost))
		    {
			open_stream(&streamhost);
		    }
		    shutdown (streamhost.sock, SHUT_RDWR);
		}
		else
		{
		    ret = -1;
		}
	    }
	    free (url);
	    break;
	}
	
        default:
	{
	    fprintf (stderr,
		     "Usage error:  Too many or unrecognized arguments.\n\n");
	    usage();
	    break;
	}
    }
    
    return (ret);
}


int parse_url (char *url, struct hostdata *streamhost)
{
    int ret = 0;
    char *proto, *orig, *p;

    
    orig = strdup(url);
    proto = strtok(url, "/");
    streamhost->hostname = strtok(NULL, "/");
    
    p = strstr(orig, streamhost->hostname);
    p += strlen(streamhost->hostname) * sizeof(char);

    if (strlen(p))
    {
	streamhost->path = strtok(NULL, " ");
    }
    
    if ((p = strchr(streamhost->hostname, ':')) != NULL)
    {
	*p = 0;
	p++;
	streamhost->port = atoi(p);
    }
    else
    {
	streamhost->port = 80;
    }
    
    if (!strncasecmp(proto, (const char *)"http:", 5) 
	&& strlen(streamhost->hostname) > 0) 
    {
	ret = 1;
    }
    else
    {
	if (!strncasecmp(proto, (const char *)"https:", 6))
	{
	    fprintf (stderr,
		     "Usage error:  Streamconv does not support the HTTPS protocol\n\n");
	}
	else
	{
	    fprintf (stderr,
		     "Usage error:  '%s' does not look like a valid HTTP URL\n\n",
		     orig);
	}
	usage();
	ret = -1;
    }
    
    free (orig);
    return (ret);
}


int connect_host (struct hostdata *streamhost)
{
    struct hostent *host;
    int ret = 0;
    
    if (host = gethostbyname(streamhost->hostname))
    {
	struct protoent *proto;
	
	proto = getprotobyname("tcp");
	if (streamhost->sock = socket(PF_INET, SOCK_STREAM, proto->p_proto))
	{
	    struct sockaddr_in saddr;
	    int c;
	    
	    bzero (&saddr, sizeof(saddr));
	    saddr.sin_family = AF_INET;
	    bcopy (host->h_addr_list[0], &saddr.sin_addr, host->h_length);
	    saddr.sin_port = htons(streamhost->port);

	    if ((c = connect(streamhost->sock, (struct sockaddr *) &saddr, sizeof(saddr))) == 0)
	    {
		fcntl(streamhost->sock, F_SETFL, FNDELAY | O_NDELAY | O_NONBLOCK);
		ret = 1;
	    }
	    else
	    {
		fprintf (stderr, "Connection failed!\n");
	    }
	}	    
    }
    else
    {
	/* If host is NULL, gethostbyname() failed.  We could check for
	 * TRY_AGAIN and retry, but we're not going to do so at this time.
	 */
	fprintf (stderr, "Error:  Unable to resolve host %s\n", streamhost->hostname);
    }
    
    return (ret);
}


int get_stream_port(struct hostdata *streamhost)
{
    char buf[BUFSIZE];
    int got_playlist = 0;
    int got_stream = 0;
    int bytes;
    fd_set rmask, wmask, xmask;


    bzero(buf, BUFSIZE);
    if (streamhost->path != NULL)
    {
	sprintf(buf,
		"GET /%s HTTP/1.1\nHost: %s\n\n", 
		streamhost->path,
		streamhost->hostname);
    }
    else
    {
	sprintf(buf,
		"GET / HTTP/1.1\nHost: %s\n\n",
		streamhost->hostname);
    }
    
    write(streamhost->sock, buf, strlen(buf));

    bzero(buf, BUFSIZE);
    
    FD_ZERO(&rmask);
    FD_ZERO(&wmask);
    FD_ZERO(&xmask);
    FD_SET(streamhost->sock, &rmask);

    while (!(got_playlist || strstr(buf,"</html>")))
    {
	select(streamhost->sock+1, &rmask, &wmask, &xmask, NULL);
    
	bytes = read(streamhost->sock, buf, BUFSIZE - 1);
	if (bytes)
	{
	    char *p;
	    
	    if ((p = strstr(buf, "[playlist]")) != NULL)
	    {
		char *l;
		
		got_playlist = 1;
		p = strdup(strstr(buf, "[playlist]"));
		l = strtok(p, "\n");
		while ((l = strtok(NULL, "\n")) != NULL && !got_stream)
		{
		    if (!strncasecmp(l,"file",4))
		    {
			char *d, *h;
			
			got_stream = 1;
			strtok(l, "/");
			d = strtok(NULL, "/");
			h = strtok(d, ":");
			p = strtok(NULL, ":");
			streamhost->stream_host = strdup(h);
			streamhost->stream_port = atoi(p); 
		    }
		}
	    }
	}
    }
    
    fprintf(stderr, "Opening stream from %s port %d\n", streamhost->stream_host, streamhost->stream_port);
    return(got_stream);
}


void open_stream(struct hostdata *streamhost)
{
    char buf[BUFSIZE], ch[2], *p;
    struct hostent *host;
    struct sockaddr_in saddr;
    struct protoent *proto;
    uint32_t a[4];
    int c, bytes, stream, btotal, bhead;
    fd_set rmask, wmask, xmask;

    
    bzero (&saddr, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(streamhost->stream_port);
    proto = getprotobyname("tcp");

    if ((c = sscanf(streamhost->stream_host, "%u.%u.%u.%u", &a[0], &a[1], &a[2], &a[3])) > 0)
    {
	inet_aton(streamhost->stream_host, &(saddr.sin_addr));
    }
    else
    {
	host = gethostbyname(streamhost->stream_host);
	bcopy (host->h_addr_list[0], &saddr.sin_addr, host->h_length);
    }
        
    if (stream = socket(PF_INET, SOCK_STREAM, proto->p_proto))
    {
	if ((c = connect(stream, (struct sockaddr *) &saddr, sizeof(saddr))) == 0)
	{
	    fcntl(stream, F_SETFL, FNDELAY | O_NDELAY | O_NONBLOCK);
	}
	else
	{
	    fprintf (stderr, "Stream connection failed!\n");
	}
	
	sprintf(buf,
		"GET / HTTP/1.1\nHost: %s\nIcy-MetaData:0\n\n", 
		streamhost->hostname);
	write(stream, buf, strlen(buf));
	
	bzero(buf, BUFSIZE);
	
	FD_ZERO(&rmask);
	FD_ZERO(&wmask);
	FD_ZERO(&xmask);
	FD_SET(stream, &rmask);
	
	streamhost->metaint = 0;
	select(stream+1, &rmask, &wmask, &xmask, NULL);
	
	bzero(ch, 2);
	bzero(buf, BUFSIZE);
	p = buf;
	btotal = bhead = 0;
	while (streamhost->metaint == 0)
	{
	    if ((bytes = read(stream, buf, BUFSIZE-1)) > 0)
	    {
		btotal += bytes;
		char *b = buf;
		while ((p = strstr(b, "\n")) != NULL)
		{
		    *p = 0;
		    bhead += strlen(b) + 1;
		    if (strlen(b) == 1)
		    {
			streamhost->metaint = 1;
		    }
		    b = ++p;
		}
	    }
	}
	
	p = buf;
	p += bhead * sizeof(char);
	c = btotal - bhead;
	
	fwrite (p, sizeof(char), c, stdout);
	btotal = c;
	
	while (1)
	{
	    if ((bytes = read(stream, buf, BUFSIZE)) > 0)
	    {
		fwrite (buf, sizeof(char), bytes, stdout);
	    }
	}
		
	shutdown (stream, SHUT_RDWR);
    }
    return;
}
    

void usage(void)
{
    fprintf (stderr, "Streamconv:   Extract an MP3 stream from a Shoutcast stream\n");
    fprintf (stderr, "Usage:        streamconv <HTTP URL> | <MP3 player or other use>\n\n");
    fprintf (stderr, "Streamconv opens a Shoutcast stream from a playlist URL specified\n");
    fprintf (stderr, "as the first and only command line argument, strips out the\n");
    fprintf (stderr, "Shoutcast metadata, and outputs the cleaned MP3 audio stream on\n");
    fprintf (stderr, "stdout.\n\n");
    
    return;
}
